﻿Imports Microsoft.VisualBasic
Imports System
Namespace RadialNetworkTrain
    Friend Class RadialTrainProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin radial basis function (RBF) network training demo" & vbCrLf)
            Console.WriteLine("Goal is to train an RBF network on iris flower data to predict")
            Console.WriteLine("species from sepal length and width, and petal length and width.")

            ' (0,0,1) = Iris setosa; (0,1,0) = Iris versicolor; (1,0,0) = Iris virginica
            ' program assumes x-data precedes y-data

            Dim allData(29)() As Double
            allData(0) = New Double() {-0.784, 1.255, -1.332, -1.306, 0, 0, 1}
            allData(1) = New Double() {-0.995, -0.109, -1.332, -1.306, 0, 0, 1}
            allData(2) = New Double() {-1.206, 0.436, -1.386, -1.306, 0, 0, 1}
            allData(3) = New Double() {-1.312, 0.164, -1.278, -1.306, 0, 0, 1}
            allData(4) = New Double() {-0.89, 1.528, -1.332, -1.306, 0, 0, 1}
            allData(5) = New Double() {-0.468, 2.346, -1.17, -1.048, 0, 0, 1}
            allData(6) = New Double() {-1.312, 0.982, -1.332, -1.177, 0, 0, 1}
            allData(7) = New Double() {-0.89, 0.982, -1.278, -1.306, 0, 0, 1}
            allData(8) = New Double() {-1.523, -0.382, -1.332, -1.306, 0, 0, 1}
            allData(9) = New Double() {-0.995, 0.164, -1.278, -1.435, 0, 0, 1}

            allData(10) = New Double() {1.22, 0.436, 0.452, 0.241, 0, 1, 0}
            allData(11) = New Double() {0.587, 0.436, 0.344, 0.37, 0, 1, 0}
            allData(12) = New Double() {1.115, 0.164, 0.56, 0.37, 0, 1, 0}
            allData(13) = New Double() {-0.362, -2.019, 0.074, 0.112, 0, 1, 0}
            allData(14) = New Double() {0.693, -0.655, 0.398, 0.37, 0, 1, 0}
            allData(15) = New Double() {-0.151, -0.655, 0.344, 0.112, 0, 1, 0}
            allData(16) = New Double() {0.482, 0.709, 0.452, 0.498, 0, 1, 0}
            allData(17) = New Double() {-0.995, -1.746, -0.305, -0.275, 0, 1, 0}
            allData(18) = New Double() {0.798, -0.382, 0.398, 0.112, 0, 1, 0}
            allData(19) = New Double() {-0.679, -0.927, 0.02, 0.241, 0, 1, 0}

            allData(20) = New Double() {0.482, 0.709, 1.155, 1.659, 1, 0, 0}
            allData(21) = New Double() {-0.046, -0.927, 0.669, 0.885, 1, 0, 0}
            allData(22) = New Double() {1.326, -0.109, 1.101, 1.143, 1, 0, 0}
            allData(23) = New Double() {0.482, -0.382, 0.939, 0.756, 1, 0, 0}
            allData(24) = New Double() {0.693, -0.109, 1.047, 1.272, 1, 0, 0}
            allData(25) = New Double() {1.853, -0.109, 1.479, 1.143, 1, 0, 0}
            allData(26) = New Double() {-0.995, -1.473, 0.344, 0.627, 1, 0, 0}
            allData(27) = New Double() {1.537, -0.382, 1.317, 0.756, 1, 0, 0}
            allData(28) = New Double() {0.904, -1.473, 1.047, 0.756, 1, 0, 0}
            allData(29) = New Double() {1.431, 1.528, 1.209, 1.659, 1, 0, 0}

            Console.WriteLine(vbCrLf & "First four and last line of normalized, encoded input data:" & vbCrLf)
            Helpers.ShowMatrix(allData, 4, 3, True, True)

            Console.WriteLine(vbCrLf & "Splitting data into 80%-20% train and test sets")
            Dim trainData()() As Double = Nothing
            Dim testData()() As Double = Nothing
            Dim seed As Integer = 8 ' gives a good demo
            GetTrainTest(allData, seed, trainData, testData) ' 80-20 hold-out

            Console.WriteLine(vbCrLf & "Training data: " & vbCrLf)
            Helpers.ShowMatrix(trainData, trainData.Length, 3, True, False)
            Console.WriteLine(vbCrLf & "Test data:" & vbCrLf)
            Helpers.ShowMatrix(testData, testData.Length, 3, True, False)

            Console.WriteLine(vbCrLf & "Creating a 4-5-3 radial basis function network")
            Dim numInput As Integer = 4
            Dim numHidden As Integer = 5
            Dim numOutput As Integer = 3
            Dim rn As New RadialNetwork(numInput, numHidden, numOutput)

            Console.WriteLine(vbCrLf & "Beginning RBF training" & vbCrLf)
            Dim maxIterations As Integer = 100 ' max for PSO
            Dim bestWeights() As Double = rn.Train(trainData, maxIterations)

            Console.WriteLine(vbCrLf & "Evaluating result RBF classification accuracy on the test data")
            rn.SetWeights(bestWeights)

            Dim acc As Double = rn.Accuracy(testData)
            Console.WriteLine("Classification accuracy = " & acc.ToString("F4"))

            Console.WriteLine(vbCrLf & "End RBF network training demo" & vbCrLf)
            Console.ReadLine()

        End Sub ' Main()

        Private Shared Sub GetTrainTest(ByVal allData()() As Double, ByVal seed As Integer, <System.Runtime.InteropServices.Out()> ByRef trainData()() As Double, <System.Runtime.InteropServices.Out()> ByRef testData()() As Double)
            ' 80-20 hold-out validation
            Dim allIndices(allData.Length - 1) As Integer
            For i As Integer = 0 To allIndices.Length - 1
                allIndices(i) = i
            Next i

            Dim rnd As New Random(seed)
            For i As Integer = 0 To allIndices.Length - 1 ' shuffle indices
                Dim r As Integer = rnd.Next(i, allIndices.Length)
                Dim tmp As Integer = allIndices(r)
                allIndices(r) = allIndices(i)
                allIndices(i) = tmp
            Next i

            Dim numTrain As Integer = CInt(Fix(0.8 * allData.Length))
            Dim numTest As Integer = allData.Length - numTrain

            trainData = New Double(numTrain - 1)() {}
            testData = New Double(numTest - 1)() {}

            Dim j As Integer = 0
            For i As Integer = 0 To numTrain - 1
                trainData(i) = allData(allIndices(j))
                j += 1
            Next i
            For i As Integer = 0 To numTest - 1
                testData(i) = allData(allIndices(j))
                j += 1
            Next i

        End Sub ' GetTrainTest



    End Class ' class Program


    Public Class RadialNetwork
        Private Shared rnd As Random = Nothing
        Private numInput As Integer
        Private numHidden As Integer
        Private numOutput As Integer
        Private inputs() As Double
        Private centroids()() As Double
        Private widths() As Double
        Private hoWeights()() As Double
        Private oBiases() As Double
        Private outputs() As Double

        Public Sub New(ByVal numInput As Integer, ByVal numHidden As Integer, ByVal numOutput As Integer)
            rnd = New Random(0)
            Me.numInput = numInput
            Me.numHidden = numHidden
            Me.numOutput = numOutput
            Me.inputs = New Double(numInput - 1) {}
            Me.centroids = MakeMatrix(numHidden, numInput)
            Me.widths = New Double(numHidden - 1) {}
            Me.hoWeights = MakeMatrix(numHidden, numOutput)
            Me.oBiases = New Double(numOutput - 1) {}
            Me.outputs = New Double(numOutput - 1) {}
        End Sub ' ctor

        Private Shared Function MakeMatrix(ByVal rows As Integer, ByVal cols As Integer) As Double()() ' helper for ctor
            Dim result(rows - 1)() As Double
            For r As Integer = 0 To rows - 1
                result(r) = New Double(cols - 1) {}
            Next r
            Return result
        End Function

        ' -- methods related to getting and setting centroids, widths, weights, bias values ----------

        Public Sub SetWeights(ByVal weights() As Double)
            ' this.hoWeights has numHidden row and numOutput cols
            ' this.oBiases has numOutput values
            If weights.Length <> (numHidden * numOutput) + numOutput Then
                Throw New Exception("Bad weights length in SetWeights")
            End If
            Dim k As Integer = 0 ' ptr into weights
            For i As Integer = 0 To numHidden - 1
                For j As Integer = 0 To numOutput - 1
                    Me.hoWeights(i)(j) = weights(k)
                    k += 1
                Next j
            Next i
            For i As Integer = 0 To numOutput - 1
                Me.oBiases(i) = weights(k)
                k += 1
            Next i
        End Sub

        Public Function GetWeights() As Double()
            Dim result(numHidden * numOutput + numOutput - 1) As Double
            Dim k As Integer = 0
            For i As Integer = 0 To numHidden - 1
                For j As Integer = 0 To numOutput - 1
                    result(k) = Me.hoWeights(i)(j)
                    k += 1
                Next j
            Next i
            For i As Integer = 0 To numOutput - 1
                result(k) = Me.oBiases(i)
                k += 1
            Next i
            Return result
        End Function

        ' put GetAllWeights() and SetAllWeights() here to fetch and set
        ' centroid, width, weight, and bias values as a group

        ' -- methods related to training error and test classification accuracy ----------------------

        Private Function MeanSquaredError(ByVal trainData()() As Double, ByVal weights() As Double) As Double
            ' assumes that centroids and widths have been set!
            Me.SetWeights(weights) ' copy the weights to valuate in

            Dim xValues(numInput - 1) As Double ' inputs
            Dim tValues(numOutput - 1) As Double ' targets
            Dim sumSquaredError As Double = 0.0
            For i As Integer = 0 To trainData.Length - 1 ' walk through each trainingb data item
                ' following assumes data has all x-values first, followed by y-values!
                Array.Copy(trainData(i), xValues, numInput) ' extract inputs
                Array.Copy(trainData(i), numInput, tValues, 0, numOutput) ' extract targets
                Dim yValues() As Double = Me.ComputeOutputs(xValues) ' compute the outputs using centroids, widths, weights, bias values
                For j As Integer = 0 To yValues.Length - 1
                    sumSquaredError += ((yValues(j) - tValues(j)) * (yValues(j) - tValues(j)))
                Next j
            Next i
            Return sumSquaredError / trainData.Length
        End Function

        ' consider MeanCrossEntropy() here as an alternative to MSE

        Public Function Accuracy(ByVal testData()() As Double) As Double
            ' percentage correct using winner-takes all
            Dim numCorrect As Integer = 0
            Dim numWrong As Integer = 0
            Dim xValues(numInput - 1) As Double ' inputs
            Dim tValues(numOutput - 1) As Double ' targets
            Dim yValues() As Double ' computed Y

            For i As Integer = 0 To testData.Length - 1
                Array.Copy(testData(i), xValues, numInput) ' parse test data into x-values and t-values
                Array.Copy(testData(i), numInput, tValues, 0, numOutput)
                yValues = Me.ComputeOutputs(xValues)
                Dim maxIndex As Integer = RadialNetwork.MaxIndex(yValues) ' which cell in yValues has largest value?
                If tValues(maxIndex) = 1.0 Then ' ugly. consider AreEqual(double x, double y)
                    numCorrect += 1
                Else
                    numWrong += 1
                End If
            Next i
            Return (numCorrect * 1.0) / (numCorrect + numWrong) ' ugly 2 - check for divide by zero
        End Function

        Private Shared Function MaxIndex(ByVal vector() As Double) As Integer ' helper for Accuracy()
            ' index of largest value
            Dim bigIndex As Integer = 0
            Dim biggestVal As Double = vector(0)
            For i As Integer = 0 To vector.Length - 1
                If vector(i) > biggestVal Then
                    biggestVal = vector(i)
                    bigIndex = i
                End If
            Next i
            Return bigIndex
        End Function

        ' -- methods related to RBF network input-output mechanism -----------------------------------

        Public Function ComputeOutputs(ByVal xValues() As Double) As Double()
            ' use centroids, widths, weights and input xValues to compute, store, and return numOutputs output values
            Array.Copy(xValues, Me.inputs, xValues.Length) ' place data inputs into RBF net inputs

            Dim hOutputs(numHidden - 1) As Double ' hidden node outputs
            For j As Integer = 0 To numHidden - 1 ' each hidden node
                Dim d As Double = EuclideanDist(inputs, centroids(j), inputs.Length) ' could use a 'distSquared' approach
                'Console.WriteLine("\nHidden[" + j + "] distance = " + d.ToString("F4"));
                Dim r As Double = -1.0 * (d * d) / (2 * widths(j) * widths(j))
                Dim g As Double = Math.Exp(r)
                'Console.WriteLine("Hidden[" + j + "] output = " + g.ToString("F4"));
                hOutputs(j) = g
            Next j

            Dim tempResults(numOutput - 1) As Double

            For k As Integer = 0 To numOutput - 1
                For j As Integer = 0 To numHidden - 1
                    tempResults(k) += (hOutputs(j) * hoWeights(j)(k)) ' accumulate
                Next j
            Next k

            For k As Integer = 0 To numOutput - 1
                tempResults(k) += oBiases(k) ' add biases
            Next k

            Dim finalOutputs() As Double = Softmax(tempResults) ' scale the raw output so values sum to 1.0

            'Console.WriteLine("outputs:");
            'Helpers.ShowVector(finalOutputs, 3, finalOutputs.Length, true);
            'Console.ReadLine();

            Array.Copy(finalOutputs, Me.outputs, finalOutputs.Length) ' transfer computed outputs to RBF net outputs

            Dim returnResult(numOutput - 1) As Double ' also return computed outputs for convenience
            Array.Copy(finalOutputs, returnResult, outputs.Length)
            Return returnResult
        End Function ' ComputeOutputs



        Private Shared Function Softmax(ByVal rawOutputs() As Double) As Double()
            ' helper for ComputeOutputs
            ' does all output nodes at once so scale doesn't have to be re-computed each time
            ' determine max output sum
            Dim max As Double = rawOutputs(0)
            For i As Integer = 0 To rawOutputs.Length - 1
                If rawOutputs(i) > max Then
                    max = rawOutputs(i)
                End If
            Next i

            ' determine scaling factor -- sum of exp(each val - max)
            Dim scale As Double = 0.0
            For i As Integer = 0 To rawOutputs.Length - 1
                scale += Math.Exp(rawOutputs(i) - max)
            Next i

            Dim result(rawOutputs.Length - 1) As Double
            For i As Integer = 0 To rawOutputs.Length - 1
                result(i) = Math.Exp(rawOutputs(i) - max) / scale
            Next i

            Return result ' now scaled so that all values sum to 1.0
        End Function

        ' -- methods related to training: DoCentroids, DistinctIndices, AvgAbsDist, DoWidths,
        '    DoWeights, Train, Shuffle ----------------------------------------------------------

        Private Sub DoCentroids(ByVal trainData()() As Double)
            ' centroids are representative inputs that are relatively different
            ' compute centroids using the x-vaue of training data
            ' store into this.centroids
            Dim numAttempts As Integer = trainData.Length
            Dim goodIndices(numHidden - 1) As Integer ' need one centroid for each hidden node
            Dim maxAvgDistance As Double = Double.MinValue ' largest average distance for a set of candidate indices
            For i As Integer = 0 To numAttempts - 1
                Dim randomIndices() As Integer = DistinctIndices(numHidden, trainData.Length) ' candidate indices
                Dim sumDists As Double = 0.0 ' sum of distances between adjacent candidates (not all candiates)
                For j As Integer = 0 To randomIndices.Length - 2 ' adjacent pairs only
                    Dim firstIndex As Integer = randomIndices(j)
                    Dim secondIndex As Integer = randomIndices(j + 1)
                    sumDists += AvgAbsDist(trainData(firstIndex), trainData(secondIndex), numInput) ' just the input terms
                Next j

                Dim estAvgDist As Double = sumDists / numInput ' estimated average distance for curr candidates
                If estAvgDist > maxAvgDistance Then ' curr candidates are far apart
                    maxAvgDistance = estAvgDist
                    Array.Copy(randomIndices, goodIndices, randomIndices.Length) ' save curr candidates
                End If
            Next i ' now try a new set of candidates

            Console.WriteLine("The indices (into training data) of the centroids are:")
            Helpers.ShowVector(goodIndices, goodIndices.Length, True)

            ' store copies of x-vales of data pointed to by good indices into this.centroids
            For i As Integer = 0 To numHidden - 1
                Dim idx As Integer = goodIndices(i) ' idx points to trainData
                For j As Integer = 0 To numInput - 1
                    Me.centroids(i)(j) = trainData(idx)(j) ' make a copy of values
                Next j
            Next i
        End Sub ' DoCentroids

        Private Shared Function AvgAbsDist(ByVal v1() As Double, ByVal v2() As Double, ByVal numTerms As Integer) As Double
            ' average absolute difference distance between two vectors, first numTerms only
            ' helper for computing centroids
            If v1.Length <> v2.Length Then
                Throw New Exception("Vector lengths not equal in AvgAbsDist()")
            End If
            Dim sum As Double = 0.0
            For i As Integer = 0 To numTerms - 1
                Dim delta As Double = Math.Abs(v1(i) - v2(i))
                sum += delta
            Next i
            Return sum / numTerms
        End Function

        Private Function DistinctIndices(ByVal n As Integer, ByVal range As Integer) As Integer()
            ' helper for ComputeCentroids()
            ' generate n distinct numbers in [0, range-1] using reservoir sampling
            ' assumes rnd exists
            Dim result(n - 1) As Integer
            For i As Integer = 0 To n - 1
                result(i) = i
            Next i

            For t As Integer = n To range - 1
                Dim m As Integer = rnd.Next(0, t + 1)
                If m < n Then
                    result(m) = t
                End If
            Next t
            Return result
        End Function

        Private Sub DoWidths(ByVal centroids()() As Double)
            ' compute widths based on centroids, store into this.widths
            ' note the centroids parameter could be omitted - the intent is to make relationship clear
            ' this version uses a common width which is the average dist between all centroids
            Dim sumOfDists As Double = 0.0
            Dim ct As Integer = 0 ' could calculate number pairs instead
            For i As Integer = 0 To centroids.Length - 2
                For j As Integer = i + 1 To centroids.Length - 1
                    Dim dist As Double = EuclideanDist(centroids(i), centroids(j), centroids(i).Length)
                    sumOfDists += dist
                    ct += 1
                Next j
            Next i
            Dim avgDist As Double = sumOfDists / ct
            Dim width As Double = avgDist

            Console.WriteLine("The common width is: " & width.ToString("F4"))

            For i As Integer = 0 To Me.widths.Length - 1 ' all widths the same
                Me.widths(i) = width
            Next i
        End Sub

         'Private Sub DoWidths(ByVal centroids()() As Double)
        '' this version uses a common width which is dmax / sqrt(2*numHidden)
        'Dim maxDist As Double = Double.MinValue
        '    For i As Integer = 0 To centroids.Length - 2
        '        For j As Integer = i + 1 To centroids.Length - 1
        'Dim dist As Double = EuclideanDist(centroids(i), centroids(j), centroids(i).Length)
        '            If dist > maxDist Then
        '                maxDist = dist
        '            End If
        '        Next j
        '    Next i

        'Dim width As Double = maxDist / Math.Sqrt(2.0 * numHidden)
        '    Console.WriteLine("The common width is: " & width.ToString("F4"))
        '    For i As Integer = 0 To Me.widths.Length - 1 ' all widths the same
        '        Me.widths(i) = width
        '    Next i
        'End Sub

        Private Function DoWeights(ByVal trainData()() As Double, ByVal maxIterations As Integer) As Double()
            ' use PSO to find weights and bias values that produce a RBF network
            ' that best matches training data
            Dim numberParticles As Integer = trainData.Length \ 3

            Dim [Dim] As Integer = (numHidden * numOutput) + numOutput ' dimensions is num weights + num biases
            Dim minX As Double = -10.0 ' implicitly assumes data has been normalizzed
            Dim maxX As Double = 10.0
            Dim minV As Double = minX
            Dim maxV As Double = maxX
            Dim swarm(numberParticles - 1) As Particle
	  [Dim] bestGlobalPosition() As Double = New Double([Dim] - 1){} ' best solution found by any particle in the swarm. implicit initialization to all 0.0
            Dim smallesttGlobalError As Double = Double.MaxValue ' smaller values better

            ' initialize swarm
            For i As Integer = 0 To swarm.Length - 1 ' initialize each Particle in the swarm
		[Dim] randomPosition() As Double = New Double([Dim] - 1){}
                For j As Integer = 0 To randomPosition.Length - 1
                    Dim lo As Double = minX
                    Dim hi As Double = maxX
                    randomPosition(j) = (hi - lo) * rnd.NextDouble() + lo
                Next j

                Dim err As Double = MeanSquaredError(trainData, randomPosition) ' error associated with the random position/solution
		[Dim] randomVelocity() As Double = New Double([Dim] - 1){}

                For j As Integer = 0 To randomVelocity.Length - 1
                    Dim lo As Double = -1.0 * Math.Abs(maxV - minV)
                    Dim hi As Double = Math.Abs(maxV - minV)
                    randomVelocity(j) = (hi - lo) * rnd.NextDouble() + lo
                Next j
                swarm(i) = New Particle(randomPosition, err, randomVelocity, randomPosition, err)

                ' does current Particle have global best position/solution?
                If swarm(i).error < smallesttGlobalError Then
                    smallesttGlobalError = swarm(i).error
                    swarm(i).position.CopyTo(bestGlobalPosition, 0)
                End If
            Next i ' initialization

            ' main PSO algorithm
            ' compute new velocity -> compute new position -> check if new best

            Dim w As Double = 0.729 ' inertia weight
            Dim c1 As Double = 1.49445 ' cognitive/local weight
            Dim c2 As Double = 1.49445 ' social/global weight
            Dim r1, r2 As Double ' cognitive and social randomizations

            Dim sequence(numberParticles - 1) As Integer ' process particles in random order
            For i As Integer = 0 To sequence.Length - 1
                sequence(i) = i
            Next i

            Dim iteration As Integer = 0
            Do While iteration < maxIterations
                If smallesttGlobalError < 0.06 Then ' early exit (MSE)
                    Exit Do
                End If

		[Dim] newVelocity() As Double = New Double([Dim] - 1){} ' step 1
		[Dim] newPosition() As Double = New Double([Dim] - 1){} ' step 2
                Dim newError As Double ' step 3

                Shuffle(sequence) ' move particles in random sequence

                For pi As Integer = 0 To swarm.Length - 1 ' each Particle (index)
                    Dim i As Integer = sequence(pi)
                    Dim currP As Particle = swarm(i) ' for coding convenience

                    ' 1. compute new velocity
                    For j As Integer = 0 To currP.velocity.Length - 1 ' each x value of the velocity
                        r1 = rnd.NextDouble()
                        r2 = rnd.NextDouble()

                        ' velocity depends on old velocity, best position of parrticle, and 
                        ' best position of any particle
                        newVelocity(j) = (w * currP.velocity(j)) + (c1 * r1 * (currP.bestPosition(j) - currP.position(j))) + (c2 * r2 * (bestGlobalPosition(j) - currP.position(j)))

                        If newVelocity(j) < minV Then
                            newVelocity(j) = minV
                        ElseIf newVelocity(j) > maxV Then
                            newVelocity(j) = maxV ' crude way to keep velocity in range
                        End If
                    Next j

                    newVelocity.CopyTo(currP.velocity, 0)

                    ' 2. use new velocity to compute new position
                    For j As Integer = 0 To currP.position.Length - 1
                        newPosition(j) = currP.position(j) + newVelocity(j) ' compute new position
                        If newPosition(j) < minX Then
                            newPosition(j) = minX
                        ElseIf newPosition(j) > maxX Then
                            newPosition(j) = maxX
                        End If
                    Next j

                    newPosition.CopyTo(currP.position, 0)

                    ' 3. use new position to compute new error
                    ' consider cross-entropy error instead of MSE
                    newError = MeanSquaredError(trainData, newPosition) ' makes next check a bit cleaner
                    currP.error = newError

                    If newError < currP.smallestError Then ' new particle best?
                        newPosition.CopyTo(currP.bestPosition, 0)
                        currP.smallestError = newError
                    End If

                    If newError < smallesttGlobalError Then ' new global best?
                        newPosition.CopyTo(bestGlobalPosition, 0)
                        smallesttGlobalError = newError
                    End If

                    ' consider using weight decay, particle death here

                Next pi ' each Particle

                iteration += 1

            Loop ' while (main PSO processing loop)

            'Console.WriteLine(vbCrLf & vbCrLf & "Final training MSE = " & smallesttGlobalError.ToString("F4")

            ' copy best weights found into RBF network, and also return them
            Me.SetWeights(bestGlobalPosition)
            Dim returnResult((numHidden * numOutput) + numOutput - 1) As Double
            Array.Copy(bestGlobalPosition, returnResult, bestGlobalPosition.Length)

            Console.WriteLine("The best weights and bias values found are:" & vbCrLf)
            Helpers.ShowVector(bestGlobalPosition, 3, 10, True)
            Return returnResult
        End Function ' DoWeights

        Private Shared Sub Shuffle(ByVal sequence() As Integer)
            ' helper for DoWeights to process particles in random order
            For i As Integer = 0 To sequence.Length - 1
                Dim r As Integer = rnd.Next(i, sequence.Length)
                Dim tmp As Integer = sequence(r)
                sequence(r) = sequence(i)
                sequence(i) = tmp
            Next i
        End Sub

        Public Function Train(ByVal trainData()() As Double, ByVal maxIterations As Integer) As Double()
            Console.WriteLine(vbCrLf & "1. Computing " & numHidden & " centroids")
            DoCentroids(trainData) ' find representative data, store their x-values into this.centroids

            Console.WriteLine(vbCrLf & "2. Computing a common width for each hidden node")
            DoWidths(Me.centroids) ' measure of how far apart centroids are

            Dim numWts As Integer = (numHidden * numOutput) + numOutput
            Console.WriteLine(vbCrLf & "3. Determining " & numWts & " weights and bias values using PSO algorithm")
            Dim bestWeights() As Double = DoWeights(trainData, maxIterations) ' use PSO to find weights that best (lowest MSE) weights and biases

            Return bestWeights
        End Function ' Train

        ' -- The Euclidean Distance function is used by RBF ComputeOutputs and also DoWidths ----

        Private Shared Function EuclideanDist(ByVal v1() As Double, ByVal v2() As Double, ByVal numTerms As Integer) As Double
            ' Euclidean distance between two vectors, first numTerms only
            ' helper for computing RBF outputs and computing hidden node widths
            If v1.Length <> v2.Length Then
                Throw New Exception("Vector lengths not equal in EuclideanDist()")
            End If
            Dim sum As Double = 0.0
            For i As Integer = 0 To numTerms - 1
                Dim delta As Double = (v1(i) - v2(i)) * (v1(i) - v2(i))
                sum += delta
            Next i
            Return Math.Sqrt(sum)
        End Function


    End Class ' class RadialNetwork

    ' ===========================================================================

    Public Class Particle
        Public position As Double ' equivalent to x-Values and/or solution
        Public [error] As Double ' error so smaller is better
        Public velocity As Double

        Public bestPosition As Double ' best position found so far by this Particle
        Public smallestError As Double

        Public Sub New(ByVal position As Double, ByVal [error] As Double, ByVal velocity As Double, ByVal bestPosition As Double, ByVal smallestError As Double)
            Me.position = New Double(position.Length - 1) {}
            position.CopyTo(Me.position, 0)
            Me.error = [error]
            Me.velocity = New Double(velocity.Length - 1) {}
            velocity.CopyTo(Me.velocity, 0)
            Me.bestPosition = New Double(bestPosition.Length - 1) {}
            bestPosition.CopyTo(Me.bestPosition, 0)
            Me.smallestError = smallestError
        End Sub

        'Public Overrides Function ToString As String
        'Dim s As String = ""
        '   s &= "==========================" & VBCrLf
        '  s &= "Position: "
        ' For i As Integer = 0 To Me.position.Length - 1
        '    s &= Me.position(i).ToString("F2") & " "
        'Next i
        's &= VBCrLf
        's &= "Error = " & Me.error.ToString("F4") & VBCrLf
        's &= "Velocity: "
        'For i As Integer = 0 To Me.velocity.Length - 1
        '   s &= Me.velocity(i).ToString("F2") & " "
        'Next i
        's &= VBCrLf
        's &= "Best Position: "
        'For i As Integer = 0 To Me.bestPosition.Length - 1
        '   s &= Me.bestPosition(i).ToString("F2") & " "
        'Next i
        's &= VBCrLf
        's &= "Smallest Error = " & Me.smallestError.ToString("F4") & VBCrLf
        's &= "==========================" & VBCrLf
        'Return s
        'End Function

    End Class ' class Particle

    ' ===========================================================================

    Public Class Helpers
        Public Shared Function MakeMatrix(ByVal rows As Integer, ByVal cols As Integer) As Double()()
            Dim result(rows - 1)() As Double
            For i As Integer = 0 To rows - 1
                result(i) = New Double(cols - 1) {}
            Next i
            Return result
        End Function

        Public Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal valsPerLine As Integer, ByVal blankLine As Boolean)
            For i As Integer = 0 To vector.Length - 1
                If i > 0 AndAlso i Mod valsPerLine = 0 Then ' max of 12 values per row
                    Console.WriteLine("")
                End If
                If vector(i) >= 0.0 Then
                    Console.Write(" ")
                End If
                Console.Write(vector(i).ToString("F" & decimals) & " ") ' n decimals
            Next i
            If blankLine Then
                Console.WriteLine(VBCrLf)
            End If
        End Sub

        Public Shared Sub ShowVector(ByVal vector() As Integer, ByVal valsPerLine As Integer, ByVal blankLine As Boolean)
            For i As Integer = 0 To vector.Length - 1
                If i > 0 AndAlso i Mod valsPerLine = 0 Then ' max of 12 values per row
                    Console.WriteLine("")
                End If
                If vector(i) >= 0.0 Then
                    Console.Write(" ")
                End If
                Console.Write(vector(i) & " ")
            Next i
            If blankLine Then
                Console.WriteLine(VBCrLf)
            End If
        End Sub

        Public Shared Sub ShowMatrix(ByVal matrix()() As Double, ByVal numRows As Integer, ByVal decimals As Integer, ByVal lineNumbering As Boolean, ByVal showLastLine As Boolean)
            Dim ct As Integer = 0
            If numRows = -1 Then ' if numRows == -1, show all rows
                numRows = Integer.MaxValue
            End If
            Dim i As Integer = 0
            Do While i < matrix.Length AndAlso ct < numRows
                If lineNumbering = True Then
                    Console.Write(i.ToString().PadLeft(3) & ": ")
                End If
                For j As Integer = 0 To matrix(0).Length - 1
                    If matrix(i)(j) >= 0.0 Then ' blank space instead of '+' sign
                        Console.Write(" ")
                    End If
                    Console.Write(matrix(i)(j).ToString("F" & decimals) & " ")
                Next j
                Console.WriteLine("")
                ct += 1
                i += 1
            Loop
            If showLastLine = True AndAlso numRows < matrix.Length Then
                Console.WriteLine("      ........" & VBCrLf & " ")
                Dim i As Integer = matrix.Length - 1
                Console.Write(i.ToString().PadLeft(3) & ": ")
                For j As Integer = 0 To matrix(0).Length - 1
                    If matrix(i)(j) >= 0.0 Then ' blank space instead of '+' sign
                        Console.Write(" ")
                    End If
                    Console.Write(matrix(i)(j).ToString("F" & decimals) & " ")
                Next j
            End If
            Console.WriteLine("")
        End Sub

    End Class ' class Helpers

    ' the raw data used in the demo is a 3-item subset of Fisher's Iris data set.
    '

    'Dim rawData(29) As Double '30-item subset of Iris data set (10 each class)
    'rawData(0) = New Double() { 5.1, 3.5, 1.4, 0.2, 0, 0, 1 } 'sepal length, sepal width, petal length, petal width ->
    'rawData(1) = New Double() { 4.9, 3.0, 1.4, 0.2, 0, 0, 1 } 'Iris setosa = 0 0 1, Iris versicolor = 0 1 0, Iris virginica = 1 0 0
    'rawData(2) = New Double() { 4.7, 3.2, 1.3, 0.2, 0, 0, 1 }
    'rawData(3) = New Double() { 4.6, 3.1, 1.5, 0.2, 0, 0, 1 }
    'rawData(4) = New Double() { 5.0, 3.6, 1.4, 0.2, 0, 0, 1 }
    'rawData(5) = New Double() { 5.4, 3.9, 1.7, 0.4, 0, 0, 1 }
    'rawData(6) = New Double() { 4.6, 3.4, 1.4, 0.3, 0, 0, 1 }
    'rawData(7) = New Double() { 5.0, 3.4, 1.5, 0.2, 0, 0, 1 }
    'rawData(8) = New Double() { 4.4, 2.9, 1.4, 0.2, 0, 0, 1 }
    'rawData(9) = New Double() { 4.9, 3.1, 1.5, 0.1, 0, 0, 1 }

    'rawData(10) = New Double() { 7.0, 3.2, 4.7, 1.4, 0, 1, 0 }
    'rawData(11) = New Double() { 6.4, 3.2, 4.5, 1.5, 0, 1, 0 }
    'rawData(12) = New Double() { 6.9, 3.1, 4.9, 1.5, 0, 1, 0 }
    'rawData(13) = New Double() { 5.5, 2.3, 4.0, 1.3, 0, 1, 0 }
    'rawData(14) = New Double() { 6.5, 2.8, 4.6, 1.5, 0, 1, 0 }
    'rawData(15) = New Double() { 5.7, 2.8, 4.5, 1.3, 0, 1, 0 }
    'rawData(16) = New Double() { 6.3, 3.3, 4.7, 1.6, 0, 1, 0 }
    'rawData(17) = New Double() { 4.9, 2.4, 3.3, 1.0, 0, 1, 0 }
    'rawData(18) = New Double() { 6.6, 2.9, 4.6, 1.3, 0, 1, 0 }
    'rawData(19) = New Double() { 5.2, 2.7, 3.9, 1.4, 0, 1, 0 }

    'rawData(20) = New Double() { 6.3, 3.3, 6.0, 2.5, 1, 0, 0 }
    'rawData(21) = New Double() { 5.8, 2.7, 5.1, 1.9, 1, 0, 0 }
    'rawData(22) = New Double() { 7.1, 3.0, 5.9, 2.1, 1, 0, 0 }
    'rawData(23) = New Double() { 6.3, 2.9, 5.6, 1.8, 1, 0, 0 }
    'rawData(24) = New Double() { 6.5, 3.0, 5.8, 2.2, 1, 0, 0 }
    'rawData(25) = New Double() { 7.6, 3.0, 6.6, 2.1, 1, 0, 0 }
    'rawData(26) = New Double() { 4.9, 2.5, 4.5, 1.7, 1, 0, 0 }
    'rawData(27) = New Double() { 7.3, 2.9, 6.3, 1.8, 1, 0, 0 }
    'rawData(28) = New Double() { 6.7, 2.5, 5.8, 1.8, 1, 0, 0 }
    'rawData(29) = New Double() { 7.2, 3.6, 6.1, 2.5, 1, 0, 0 }

    'Dim originalMeans() As Double = {5.843, 3.04, 3.863, 1.213} ' used to normalize original data
    'Dim originalStdDevs() As Double = {0.948, 0.367, 1.85, 0.776}

    '

End Namespace ' ns